(******************************************************************************)
(** * SRAG simulates SRAL: read step   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import Basic.
Require Import Labels Events Execution.
Require Import SRAL Wmax SRAG.
Require Import Write_lists Match.

Require Import FG.

Set Implicit Arguments.
Remove Hints plus_n_O.

Lemma GsimL_read_pick_w (st st' : State) (WF_S: wf_state st')
          (G : execution)
          t o
          (READ: forall ol', st' t ol' -> st t (o :: ol'))
          (MATCH: SGmatch st G) : 
          exists e wl' ol', 
          st t (o :: ol') /\ 
          wf_wl (extend_wl wl' e) (S (length ol')) (o :: ol') G /\ 
          consistent_wl (extend_wl wl' e) (S (length ol')) G t /\
          forall e' wl' ol' (IN: st t (o :: ol'))
            (WF_WL: wf_wl (extend_wl wl' e') (S (length ol')) (o:: ol') G)
            (CONS_WL : consistent_wl (extend_wl wl' e') (S (length ol')) G t),
            (hb G)^? e e'.
Proof.
red in MATCH; desc.
set (EE := fun e => exists wl ol', 
          st t (o :: ol') /\ 
          wf_wl (extend_wl wl e) (S (length ol')) (o :: ol') G /\ 
          consistent_wl (extend_wl wl e) (S (length ol')) G t).
assert (SOME: exists e, EE e).
{ destruct (WF_S t) as [ol OL_IN].
  destruct (SMATCH _ _ (READ ol OL_IN)) as [wl [WF_WL CONS_WL]].
  exists (wl 0), (fun n => wl (S n)), ol.
  by splits; eauto; rewrite extend_wl_wl.
}
destruct SOME as [e EEe].
assert (INe: In e G.(acts)).
by red in EEe; desc; apply EEe0; vauto; apply extend_wl_in_wl.
set (s := G.(sb)^{-1} ;; ⦗EE⦘).
assert (MIN: exists e0, s^* e e0 /\ max_elt s e0).
{ apply last_exists with (dom := G.(acts)).
  - unfold s.
    arewrite ((sb G)⁻¹ ⨾ ⦗EE⦘ ⊆ (sb G)⁻¹).
    red.
    rewrite <- transp_ct.
    rewrite ct_of_trans.
    unfolder; ins.
    eby eapply sb_irr.
    apply sb_trans.
  - ins.
    apply rt_of_trans in H.
    unfold s in H.
    hahn_rewrite wf_sbE in H.
    revert H; basic_solver.
    unfold s.
    red; unfolder; ins; desf.
    exists z; splits; eauto.
    eapply sb_trans; eauto.
}
destruct MIN as [e0 [S MAX]].
unfold s in *; clear s.
assert (EE0: EE e0).
{ apply rtE in S.
  hahn_rewrite inclusion_ct_seq_eqv_r in S.
  unfolder in S; desf.
}
clear S EEe INe e.
rename e0 into e.
unfold EE in *; clear EE.
destruct EE0 as [wl' [ol' [IN [WF_WL CONS]]]].
exists e, wl', ol'; splits; try done.
ins.
assert (A: (⦗acts_set G⦘ ⨾ same_tid ⨾ ⦗acts_set G⦘) e e').
{ clear MAX. red in WF_WL, WF_WL0; desc.
  specialize (WL_G e' (extend_wl_in_wl _ _ _)).
  specialize (WL_G0 e (extend_wl_in_wl _ _ _)).
  specialize (WL_TID 0 (Nat.lt_0_succ _)); ins.
  specialize (WL_TID0 0 (Nat.lt_0_succ _)); ins.
  unfold extend_wl in WL_TID; ins; desf.
  unfold extend_wl in WL_TID; ins; desf.
  unfold tid_opt in *.
  basic_solver 12.
}
generalize ((tid_sb WF_G) e e' A).
unfolder; ins; desf.
- basic_solver.
- right; apply sb_in_hb; done.
- exfalso; eapply MAX; basic_solver 12.
- left.
  destruct e, e'; desf.
  red in WF_WL, WF_WL0; desc.
  specialize (WL_LOC 0 (Nat.lt_0_succ _)); ins.
  specialize (WL_LOC0 0 (Nat.lt_0_succ _)); ins.
  unfold extend_wl in WL_LOC; ins; desf.
  unfold extend_wl in WL_LOC0; ins; desf.
  unfold loc_opt, loc in *; ins.
  basic_solver 12.
Qed.

Lemma GsimL_read_helper_t G (WF_G : Wf G)
 n G' t u e x v tw
 (LOC: loc e = x) (VAL: valw e = v) (TID: tid e = tw)
 (ADD : add G G' t n (Aload x v) e)
 ol wl 
 (MATCH : SGmatch_wl G t (R_opt tw x v u :: ol) wl)
 (WL_RMW : u = true -> ~ dom_rel (co G) e)
 (MIN : dom_rel (co G  ;; (hb G)^? ⨾ ⦗eq e⦘) ⊆₁ 
        dom_rel (co G  ;; (hb G)^? ⨾ ⦗eq (wl 0)⦘)):
 SGmatch_wl G' t ol (fun n => wl (S n)).
Proof.
cdes MATCH; cdes WL_WF.
assert (WL_TID' := WL_TID).
specialize (WL_TID 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_TID; ins; desf.
assert (WL_LOC' := WL_LOC).
specialize (WL_LOC 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_LOC; ins; desf.
assert (WL_VAL' := WL_VAL).
specialize (WL_VAL 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_VAL; ins; desf.
assert (WL_RMW' := WL_RMW0).
specialize (WL_RMW0 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_RMW0; ins; desf.
red; splits.
- red; splits.
  + by rewrite in_wl_monotone.
  + rewrite <- add_acts_monotone; try edone.
    by rewrite in_wl_monotone.
  + ins.
    specialize (WL_TID' (S n0) (lt_n_S _ _ LEN)); desc.
    by rewrite WL_TID'.
  + ins.
    specialize (WL_LOC' (S n0) (lt_n_S _ _ LEN)); desc.
    by rewrite WL_LOC'.
  + ins.
    specialize (WL_VAL' (S n0) (lt_n_S _ _ LEN)); desc.
    by rewrite WL_VAL'.
  + ins.
    specialize (WL_RMW' (S n0) (lt_n_S _ _ LEN)); desc; ins.
    apply WL_RMW' in RMWF.
    revert RMWF; unfolder; ins; intro; desc.
    eapply add_non_write_co with (G:=G) in H; try edone.
    2: type_solver.
    eauto.
- cdes WL_CONS.
  red; splits.
  + rewrite add_read_non_write_dom_co_hb; try edone.
    2,3: intros; type_solver.
    relsf; splits.
    rewrite in_wl_monotone; revert WL_D; basic_solver 12.
    rewrite MIN.
    unfold in_wl.
    unfolder; ins; desc; subst y.
    eapply (WL_O (wl 0)).
    unfold order_wl; unfolder; ins; eexists; splits; eauto.
    exists 0,(S n0); splits; eauto; omega.
  + rewrite add_read_non_write_co_hb; try edone.
    2,3: intros; type_solver.
    relsf; splits.
    apply irreflexive_union; splits.
    by rewrite order_wl_monotone. 
    rewrite (dom_l (order_wlE _ _)).
    rewrite in_wl_monotone, WL_G.
    unfolder; ins; apply (add_fresh ADD); desf.
Qed.

Lemma GsimL_read_helper_p G (WF_G : Wf G)
 n G' t e (ADD : add G G' t n (Aload (loc e) (valw e)) e)
 ol wl p (NEQ: p <> t) (MATCH : SGmatch_wl G p ol wl) :
 SGmatch_wl G' p ol wl.
Proof.
red in MATCH; desc.
red in WL_WF; desc.
red; splits.
- red; splits; try done.
  eby rewrite WL_G; eapply add_acts_monotone.
  unfolder; ins; intro; desc.
  eapply add_non_write_co with (G:=G) in H; try edone.
  2: type_solver.
  eapply WL_RMW; try edone.
  basic_solver.
- red in WL_CONS; desc; red; splits.
  + rewrite add_read_non_write_dom_co_hb; try edone.
    2,3: intros; type_solver.
    revert WL_D; unfold is_tid; basic_solver 12.
  + ins. rewrite add_read_non_write_co_hb; try edone.
    2,3: intros; type_solver.
    relsf; apply irreflexive_union; splits; try done.
    rewrite (dom_l (order_wlE _ _)), WL_G.
    unfolder; ins; apply (add_fresh ADD); desf.
Qed.

Lemma GsimL_read (st st' : State) (WF_S: wf_state st')
          (G : execution) (WF_G: Wf G)
          t x v (SRAL: SRAL_step st st' (Some (t, (Aload x v))))
          (MATCH: SGmatch st G) : 
          exists G', SRAG_step G G' (Some (t, (Aload x v))) /\ 
                     SGmatch st' G'.
Proof.
inversion SRAL; subst.
red in READ; subst; desf.
exploit GsimL_read_pick_w; try edone.
intro X. 
destruct X as [e [wl' [ol' [IN [WF_WL [CONS_WL MIN]]]]]].
red in WF_WL; desc.
specialize (WL_TID 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_TID; ins; desc.
specialize (WL_LOC 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_LOC; ins; desc.
specialize (WL_VAL 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_VAL; ins; desc.
specialize (WL_RMW 0 (Nat.lt_0_succ _)); ins.
unfold extend_wl in WL_RMW; ins; desc.
specialize (WL_W e (extend_wl_in_wl _ _ _)).
specialize (WL_G e (extend_wl_in_wl _ _ _)).
forward (eapply exists_add' with (w:=e) (l:= Aload x v) (t:=t)); try edone.
intro ADD; desc; exists G'; splits.
- econs; eexists _,_,e; splits; eauto; ins.
  intro. red in CONS_WL; desc.
  eapply (WL_D e); split; vauto.
  apply extend_wl_in_wl.
- red in MATCH; desc.
  red; splits; try done.
  by eapply add_preserves_wf; edone.
  subst; ins; destruct (classic (p = t)) as [X|NEQ]; try subst p.
  + apply T in IN0.
    specialize (SMATCH _ _ IN0); desc.
    eexists; eapply GsimL_read_helper_t; try edone.
    cut ((hb G)^? e (wl 0)).
    by generalize (@hb_trans G); basic_solver 12.
    by eapply MIN; eauto; rewrite extend_wl_wl; eauto; apply SMATCH.
  + apply P in IN0; try edone.
    specialize (SMATCH _ ol IN0); desc.
    eexists; eapply GsimL_read_helper_p; try edone.
Qed.