(******************************************************************************)
(** * SRAL simulates SRAG: read step   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import AuxRel.

Require Import Basic.
Require Import Filters.

Require Import Labels Events Execution.
Require Import SRAL Wmax SRAG.
Require Import Write_lists Match.

Require Import FG.

Set Implicit Arguments.
Remove Hints plus_n_O.


Lemma LsimG_read_wl_helper (G G' : execution)
  t n w l (ADD : add G G' t n l w)
  p ol wl (MATCH' : SGmatch_wl G' p ol wl)
  (WL: ~ in_wl wl (length ol) (ThreadEvent t n l))  :
SGmatch_wl G p ol wl.
Proof.
red in MATCH'; desc.
red in WL_WF; desc.
red in WL_CONS; desc.
red; unfold wf_wl,consistent_wl; unnw.
rewrite (add_co_monotone ADD), (add_hb_monotone ADD).
splits; try edone.
- rewrite (add_acts_set ADD) in WL_G.
  unfolder; intros a INWL.
  specialize (WL_G a INWL).
  unfolder in WL_G; desf.
- ins; intro.
  eapply WL_RMW; try edone.
  revert H; generalize (add_co_monotone ADD); basic_solver.
Qed.

Lemma LsimG_read_wl (G G' : execution)
  t n x v w (ADD : add G G' t n (Aload x v) w)
  p ol wl (MATCH' : SGmatch_wl G' p ol wl) :
SGmatch_wl G p ol wl.
Proof.
 eapply LsimG_read_wl_helper; try edone.
    intro INwl.
    red in MATCH'; desc.
    red in WL_WF; desc.
    specialize (WL_W _ INwl).
    desf.
Qed.

Lemma LsimG_read_wl_t (G G' : execution) (WF_G: Wf G)
  t n x v w (ADD : add G G' t n (Aload x v) w)
  ol wl (MATCH' : SGmatch_wl G' t ol wl)
  (CO : ~ dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) w)
  u (MAX: u = true -> w_max G x w):
  SGmatch_wl G t (make_r_opt w u :: ol) (extend_wl wl w).
Proof.
    red in MATCH'; desc.
    red in WL_WF; desc.
    red in WL_CONS; desc.
    red; splits; red; ins; splits; desc;
    simpl; rewrite ?in_wl_extend_wl.
    * rewrite WL_W.
      red in ADD; desc.
      basic_solver.
    * arewrite (in_wl wl (length ol) ⊆₁ 
        in_wl wl (length ol) ∩₁in_wl wl (length ol)).
      rewrite WL_W at 1.
      rewrite WL_G at 1.
      rewrite (add_acts_set ADD).
      red in ADD; desc.
      basic_solver.
  * ins; destruct n0.
    by unfold extend_wl; simpl.
    unfold extend_wl; desf.
    rewrite <- pred_Sn.
    apply WL_TID; omega.
  * ins; destruct n0.
    by unfold extend_wl; simpl.
    unfold extend_wl; desf.
    rewrite <- pred_Sn.
    apply WL_LOC; omega.
  * ins; destruct n0.
    by unfold extend_wl; simpl.
    unfold extend_wl; desf.
    rewrite <- pred_Sn.
    apply WL_VAL; omega.
  * ins; unfold extend_wl; destruct n0; desf.
    by apply dom_max_elt, MAX.
    rewrite <- pred_Sn.
    intro; eapply WL_RMW; try edone; try omega.
    revert H; generalize (add_co_monotone ADD); basic_solver.
  * relsf; splits.
    rewrite (add_co_monotone ADD), (add_hb_monotone ADD); try done.
    revert CO; basic_solver 12.
  * rewrite order_wl_extend_wl; relsf.
    apply irreflexive_union; splits.
    rewrite (add_co_monotone ADD), (add_hb_monotone ADD); done.
    unfolder; ins; desc.
    rewrite add_read_non_write_dom_co_hb in WL_D; try edone.
    2,3 : type_solver.
    eapply WL_D.
    basic_solver 12.
Qed.


Lemma LsimG_read_helper (st' : State) 
  (G G' : execution) (WF_G: Wf G)
  t x v n w 
 (ADD: add G G' t n (Aload x v) w)
  u 
  st (ST: st = fun p ol => p <> t /\ st' p ol \/
    p = t /\ exists ol', st' t ol' /\ ol = (make_r_opt w u :: ol')) : 
  SRAL_read_step_helper st st' t x v u.
Proof.
red; splits.
  + ins; desf; eauto.
  + exists (tid w); ins.
    ins; desf; right; splits; try done.
    exists ol'; splits; try done.
    unfold make_r_opt.
    red in ADD; desc; ins.
    by rewrite LOCw, VALw.
Qed.

Lemma LsimG_read_match_helper (st' : State) 
  (G G' : execution) (WF_G: Wf G)
  t x v n w 
  (ADD: add G G' t n (Aload x v) w)
  (CO: ~ dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) w) 
  (MATCH': SGmatch st' G') 
  u (MAX: u = true -> w_max G x w)
  st (ST: st = fun p ol => p <> t /\ st' p ol \/
    p = t /\ exists ol', st' t ol' /\ ol = (make_r_opt w u :: ol')) : 
  SGmatch st G.
Proof.
red; splits; try done.
- red in MATCH'; desc.
red; ins.
specialize (WF_ST t0); desc.
destruct (classic (t0 = t)); subst; eexists; eauto.
- ins. red in MATCH'.
subst; desf.
+ specialize (SMATCH p ol IN0); desc.
    eexists; eapply LsimG_read_wl; try edone.
  + specialize (SMATCH t ol' IN0); desc.
    eexists; eapply LsimG_read_wl_t; try edone.
Qed.

Lemma LsimG_read (st' : State) (WF_S: wf_state st')
          (G G' : execution) (WF_G: Wf G)
          t x v (SRAG: SRAG_step G G' (Some (t, (Aload x v))))
          (MATCH': SGmatch st' G') : 
          exists st, SRAL_step st st' (Some (t, (Aload x v))) /\ 
                     SGmatch st G.
Proof.
inversion SRAG; subst; clear SRAG; desf; ins.
eexists; splits.
eapply step_read; try done.
by eapply LsimG_read_helper; try edone; auto.
eapply LsimG_read_match_helper; try edone; auto.
Qed.
