(******************************************************************************)
(** * SRAL simulates SRAG: rmw step   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import AuxRel.

Require Import Basic.
Require Import Filters.

Require Import Labels Events Execution.
Require Import SRAL Wmax Add_max SRAG.
Require Import Write_lists Match.

Require Import FG.

Require Import LsimG_read LsimG_write.

Set Implicit Arguments.
Remove Hints plus_n_O.

Lemma LsimG_rmw_match_helper 
 (G G' : execution) (WF_G: Wf G)
 t x vr vw nrmw w
 (MAX: w_max G x w)
 (ADD: add G G' t nrmw (Armw x vr vw) w)
 nr nw Gr Grw
 (RADD: add G Gr t nr (Aload x vr) w)
 (WADD: add Gr Grw t nw (Astore x vw) w) 
 p ol wl' (MATCH':  SGmatch_wl G' p ol wl') 
 wl (WL: wl = fun m => if (excluded_middle_informative 
             (wl' m = ThreadEvent t nrmw (Armw x vr vw))) then
             (ThreadEvent t nw (Astore x vw))
             else wl' m) :
 SGmatch_wl Grw p ol wl.
Proof.
assert (WFr: Wf Gr).
by eapply add_preserves_wf; try edone.
assert (MAXr: w_max Gr x w).
by eapply w_max_add_read with (G:=G); try edone; ins.
red in MATCH'; desc.
red in WL_WF; desc.

assert (WWL: in_wl wl (length ol) ⊆₁ W).
by unfold in_wl; unfolder; ins; desf; splits; try done;
      apply WL_W; desc; red; eauto.

red; splits; red; splits; try done.
- unfold in_wl; red; ins; desf.
  by eapply add_acts_set; basic_solver.
  assert (A:acts_set G' (wl' n)).
  by apply WL_G; red; eauto.
  eapply add_acts_set; try edone; left.
  eapply add_acts_monotone; try edone.
  eapply add_acts_set in A; try edone.
  unfolder in A; desf; congruence.
- ins; desf.
  by rewrite <- WL_TID, e; ins.
  by rewrite <- WL_TID; ins.
- ins; desf.
  by rewrite <- WL_LOC, e; ins.
  by rewrite <- WL_LOC; ins.
- ins; desf.
  by rewrite <- WL_VAL, e; ins.
  by rewrite <- WL_VAL; ins.
- ins; desf.
  + apply dom_max_elt.
    eapply w_max_add_write with (G:=Gr); try edone.
  + intro.
    unfolder in H; desc.
    eapply add_max_write_co with (G:=Gr) in H; try edone; ins.
    unfolder in H; desf.
    * eapply add_non_write_co with (G:=G) in H; try edone; ins.
      eapply WL_RMW; try edone.
      exists y.
      eapply add_max_write_co with (G:=G); try edone; ins.
      basic_solver.
    * eapply add_acts_set in H; try edone.
      unfolder in H; desf; [| by rewrite <- H in H2; desf].
      eapply WL_RMW; try edone.
      exists (ThreadEvent t nrmw (Armw x vr vw)).
      eapply add_max_write_co with (G:=G) (l:=(Armw x vr vw)); try edone.
      right; basic_solver.
- rewrite add_max_write_dom_co_hb; try edone; ins.
  rewrite add_read_non_write_dom_co_hb with (G:=G); try edone; ins.
  rewrite <- !seqA.
  rewrite add_read_non_write_co_hb with (G:=G) (G':=Gr); try edone; ins.
  rewrite add_acts_set; try edone; ins.
  red in WL_CONS; desc.
  rewrite add_max_write_dom_co_hb with (G:=G) (G':=G') in WL_D; try edone; ins.
  unfolder in WL_D.
  rewrite (wf_coE WF_G).
  unfold in_wl in *; desf; unfolder; ins; desf; eauto 21.
  all: try by eapply add_fresh with (G:=Gr); try edone; eapply add_acts_set; try edone; ins; basic_solver.
  by rewrite <- H0 in H4; desf.
- rewrite add_max_write_co_hb; try edone; ins.
  arewrite (dom_rel (co Gr ⨾ (hb Gr)^? ⨾ ⦗eq w⦘)
            ∩₁ (fun _ : Event => false) ⊆₁ ∅).
  basic_solver.
  rewrite add_read_non_write_dom_co_hb with (G:=G); try edone; ins.
  rewrite <- !seqA.
  rewrite add_read_non_write_co_hb with (G:=G) (G':=Gr); try edone; ins.
  rewrite add_acts_set; try edone; ins.
  arewrite ((fun _ : Event => t = t) ⊆₁ set_full).

  arewrite ((acts_set G ∪₁ eq (ThreadEvent t nr (Aload x vr)))
              ∩₁ W ∩₁ is_loc x ⊆₁ acts_set G ∩₁ W ∩₁ is_loc x).
  by basic_solver.

  assert (AA: ~ acts_set G (ThreadEvent t nw (Astore x vw))).
  by intro; eapply add_fresh with (G:=Gr); try edone; eapply add_acts_set; try edone; ins; basic_solver.

  assert (AA1: forall n1 n2, n1 < n2 -> n2 < length ol ->
    ThreadEvent t nw (Astore x vw) <> wl' n1).
  { ins; intro X.
  forward (eapply (WL_G (wl' n1))).
  by red; eexists; splits; auto; omega.
  intro A.
  eapply add_acts_set in A; try edone; unfolder in A; desf;
  by rewrite <- X in A.
  }

  relsf.

  red in WL_CONS; desc.
  rewrite add_max_write_co_hb with (G:=G) (G':=G') in WL_O; try edone; ins.
  clear WL_D.

  rewrite !irreflexive_union; splits.
  + rewrite (wf_coE WF_G), (dom_r (wf_hbE WF_G)).
    by unfold in_wl, order_wl in *; desf; unfolder; ins; desf;
    eapply WL_O; unfolder; eexists; splits; eauto 6.
  + rewrite order_wlE, WWL; basic_solver.
  + rewrite (dom_l (wf_coE WF_G)).
    unfold in_wl, order_wl in *; desf; unfolder; ins; desf.
    all: try by eapply AA1 with (n1:=n1); eauto 21.
    all: eapply WL_O; unfolder; eexists; splits; eauto 10. 
Qed.


Lemma LsimG_rmw (st'' : State) 
          (G G' : execution) 
          t x vr vw (SRAG: SRAG_step G G' (Some (t, (Armw x vr vw))))
           (MATCH': SGmatch st'' G') :
          exists st, SRAL_step st st'' (Some (t, (Armw x vr vw))) /\ 
                     SGmatch st G.
Proof.
inversion SRAG; subst; clear SRAG; desf; ins.
specialize (MAX hahn__true_is_true).
clear CO.
rename x0 into t.
rename n into x0.

forward (eapply exists_add' with (w:=w) (l:= Aload x vr) (t:=t) (G:=G)); try edone.
all: try by red in MAX; revert MAX; basic_solver.
by ins; red in ADD; desc; auto.
intro RADD; desc.
rename G'0 into Gr.

forward (eapply exists_add' with (w:=w) (l:= Astore x vw) (t:=t) (G:=Gr)); try edone.
eby eapply add_preserves_wf.
by eapply add_acts_monotone; try edone; red in RADD; desc.
by red in RADD; desc.
by red in RADD; desc.

intro WADD; desc.
rename G'0 into Grw.

red in MATCH'; desc.

forward (apply LsimG_write with (G:=Gr) (G':=Grw)); try edone.
econs; eexists _,_,w; splits; ins; try edone.
eby eapply add_preserves_wf.
eby eapply w_max_add_read.

red; ins; splits; try edone.
eby eapply add_preserves_wf; [eapply add_preserves_wf|].
ins; specialize (SMATCH p ol IN); desc.
eexists; eapply LsimG_rmw_match_helper; try edone.

ins; desc.
rename st into st'.

forward (eapply LsimG_read_helper with (G:=G) (G':=Gr) (st':=st') (u:=true)); try edone.

ins; desc.
exists (fun (p : nat) (ol : list Opt) =>
   p <> t /\ st' p ol \/
   p = t /\
   (exists ol' : list Opt,
      st' t ol' /\ ol = make_r_opt w true :: ol')); splits.


inversion H; subst; clear H; econs; eauto.

eapply LsimG_read_match_helper with (u:=true) (st':=st'); try edone.
apply dom_max_elt; red in MAX; revert MAX; basic_solver.

Qed.
